#!/usr/bin/perl 

package DataCash::XMLDocument;

use XML::Hash;
@ISA = qw(XML::Hash);
use strict;
use Data::Dumper;
use DataCash::Config;

$DataCash::XMLDocument::VERSION = do { my @r = (q$Revision: 1.7 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };

sub new {
  
  my $package = shift;

  my %params=@_;
  my $config = undef;

  if (exists $params{configfile} && defined $params{configfile}) {
    
    $config = new DataCash::Config($params{configfile});

  }

  my $self = (defined $config) ? 
    $package->SUPER::new(config => $config) : 
      $package->SUPER::new();

  foreach my $key (keys %params) {
    if (grep /^$key$/,qw(debug configfile)) {
      $self->{$key}=$params{$key};
    }
  }
  
  $self->_useragent();

  # Create a Config object
  if (defined $self->{configfile}) {
    $self->{config} = $config;
  }
  
  return $self;

}

sub reference {

  my $self=shift;
  my $otr=shift;
  if ($self->get("Transaction.CardTxn")){
    $self->set("Transaction.TxnDetails.merchantreference",$otr);
  }
  elsif($self->get("Transaction.ChequeTxn")){
    $self->set("Transaction.TxnDetails.merchantreference",$otr);
  }
  elsif($self->get("Transaction.HistoricTxn")){
    $self->set("Transaction.HistoricTxn.reference",$otr);
  }
  else {
    die "can't set reference, until txn type is set\n";
  }
  
}

sub authentication {

  my $self=shift;
  my %details=@_;
  
  foreach my $key (keys %details) {
    if (grep /^$key$/,qw(client password)) {
      $self->set("Authentication.$key",$details{$key});
    }
  }

}

sub carddetails {
  
  my $self=shift;
  my %details=@_;
  
  foreach my $key (keys %details) {
    if (grep /^$key$/,qw(pan expirydate startdate issuenumber)) {
      $self->set("Transaction.CardTxn.Card.$key",$details{$key});
    }
  }

}

sub chequedetails {

  my $self=shift;
  my %details=@_;
  
  if (exists $details{type}) {
    $self->set("Transaction.ChequeTxn.Cheque",undef,type=>$details{type});
    delete $details{type};
  }
  
  foreach my $key (keys %details) {
    if (grep /^$key$/,qw(sortcode accountnumber chequenumber)) {
      $self->set("Transaction.ChequeTxn.Cheque.$key",$details{$key});
    }
  }

}

sub amount {

  my $self=shift;
  my %details=@_;
  
  my $amount=$details{amount};
  my $currency=$details{currency} || "GBP";
  
  #	if (eval "require DataCash::Extras") {
  #		unless (DataCash::Extras::currency($currency)) {
  #			die "Not a valid currency code";
  #		}
  #	}
  
  $self->set("Transaction.TxnDetails.amount",$amount,currency=>$currency);

}

sub method {

  my $self=shift;
  my $method=shift;
  
  if (grep /^$method$/,qw(auth pre credit refund erp cheque)) {
    if ($self->get("Transaction.CardTxn")) {
      $self->set("Transaction.CardTxn.method",$method);
    } elsif ($self->get("Transaction.ChequeTxn")) {
      $self->set("Transaction.ChequeTxn.method",$method);
    } else {
      die "Not a valid method for your class \n";
    }
  } elsif (grep /^$method$/,qw(fulfill cancel query txn_refund)) {
    $self->set("Transaction.HistoricTxn.method",$method);
  } else {
    die "Not a valid transaction method \n";
  }

}

sub cv2 {
  
  my $self=shift;
  my $cv2=shift;
  
  $self->set("Transaction.CardTxn.Card.Cv2Avs.cv2",$cv2);
  
}

sub avs {
  
  my $self=shift;
  my %details=@_;
  
  foreach my $key (keys %details) {
    if (grep /^$key$/,qw(street_address1 street_address2 street_address3 
			 street_address4 postcode policy)) {
      $self->set("Transaction.CardTxn.Card.Cv2Avs.$key",$details{$key});
    }
  }

}

sub authcode {

  my $self=shift;
  my $authcode=shift;
  if($self->get("Transaction.CardTxn")){
    $self->set("Transaction.CardTxn.authcode", $authcode);
  }elsif($self->get("Transaction.HistoricTxn")){
    $self->set("Transaction.HistoricTxn.authcode", $authcode);
  }else{
    die "Can't set authcode, set CardTxn or HistoricTxn details first\n";
  }

}

sub txntime {
  
  my $self=shift;
  my $txntime=shift;
  $self->set("Transaction.TxnDetails.txntime", $txntime);
  
}

sub txndate {

  my $self=shift;
  my $txndate=shift;
  $self->set("Transaction.TxnDetails.txndate", $txndate);

}

sub lid {
  
  my $self = shift;
  my %details = @_;
  foreach my $key (keys %details) {
    if (grep /^$key$/, qw(receiptno transactionVAT customerVATnumber supplierorderreference
			  originalorderdate discountamount transactionVATstatus)) {
      $self->set("Transaction.TxnDetails.LineItemDetail.$key", $details{$key});
    }
  }
  
}

sub lid_shipping {
  
  my $self = shift;
  my %details = @_;
  foreach my $key (keys %details) {
    if (grep /^$key$/, qw(shippingamount destinationpostalcode shipfrompostalcode
			  destinationcountrycode shippingVATrate)) {
      $self->set("Transaction.TxnDetails.LineItemDetail.Shipping.$key", $details{$key});
    }
  }
  
}

sub lid_item {
  
  my $self = shift;
  my $details = {};
  while (@_) {
    my $key = shift;
    my $value = shift;
    $details->{$key} = $value;
  }
  my $index = 0;

  if ( !defined $self->{xml}{Transaction}{TxnDetails}{LineItemDetail}{Items}{Item} ) { 
    $self->{xml}{Transaction}{TxnDetails}{LineItemDetail}{Items}{Item} = []
  }
  
  push @{$self->{xml}{Transaction}{TxnDetails}{LineItemDetail}{Items}{Item}}, $details;
  
}

sub _useragent {

  my $self=shift;
  
  ($self->{package}) = (q$Name:  $ =~ /(\d+\-\d+)/);
  $self->set("UserAgent.Libraries",undef,bundle=>"DataCash-XMLPerl",
	     version=> $self->{package});
  $self->set("UserAgent.Libraries.lib","DataCash::XMLDocument",
	     version=>$DataCash::XMLDocument::VERSION);
  $self->set("UserAgent.Libraries.lib","XML::Hash",
	     version=>$XML::Hash::VERSION);
#  if (eval "require DataCash::Extras") {
#    $self->set("UserAgent.Libraries.lib","DataCash::Extras",
#	       version=>$DataCash::Extras::VERSION);
#  }
  if (eval "require DataCash::CardInfo") {
    $self->set("UserAgent.Libraries.lib","DataCash::CardInfo",
	       version=>$DataCash::CardInfo::VERSION);
  }
  
  if (eval "require Config") {
    ## this should be some sort of a require, with necessary import schtick...
    use Config;
    $self->set("UserAgent.architecture",$Config{archname},version=>$Config{osvers});
  } else {
    $self->set("UserAgent.architecture","Config not available");
	}
  
  $self->_log(5,"set UserAgent");

}

sub setcardinfo {
  
  my $self = shift;
  
  # Make sure what we get is a DataCash::CardInfo object, and if it
  # is, do the validation.
  
  if (@_) {
    $self->{cardinfo} = shift;
    
    # Find out if the CardInfo module has its config set,
    # and if not, let's set it
    if (!defined $self->{cardinfo}->{config}) {
      $self->{cardinfo}->{config} = $self->{config};
    }
    
    if (ref($self->{cardinfo}) eq 'DataCash::CardInfo') {
      $self->_log(5, "Got a DataCash::CardInfo object to set");
      # Remember to pass it the Request object too...
      my $validcheck = $self->{cardinfo}->validation($self);
      $self->_log(5, "validcheck is: ".ref($validcheck));
      if (ref($validcheck) eq 'DataCash::XMLDocument') {
	$self->_log(2, "Failed CardInfo check:\n" . $validcheck->writexml("Response"));
	$self->{failed_cardinfo} = 'true';
	$self->{xmlresponse} = $validcheck;
	return $validcheck;
      }
      else {
	$self->_log(2,"Valid Check was a " . ref($validcheck));
	return;
      }
    }
  }
  
}

1;

########################################################################################

__END__

=head1 NAME

DataCash::XMLDocument - object which is used to create an XML document.

=head1 SYNOPSIS

  my $self = new DataCash::XMLDocument(configfile => "/home/user/datacash.conf");
  $self->set("Authentication.client", "99000000");
  $self->set("Authentication.password", "secretpassword");
  $self->set("Transaction.TxnDetails.merchantreference", "1234506789");
  $self->set("Transaction.TxnDetails.amount", "313.37");
  $self->set("Transaction.CardTxn.method", "pre");
  $self->set("Transaction.CardTxn.Card.pan", "5473000000000007");
  $self->set("Transaction.CardTxn.Card.expirydate", "11/09");

=head1 DESCRIPTION

An object-orientated module which extends <code>XML::Hash</code>.  
Includes accessor methods to some of the more frequently used elements
in XML documents that are sent to DataCash.

=head1 PUBLIC METHODS

=head2 new()

  my $self = new DataCash::XMLDocument(configfile => "/home/user/datacash.conf");

The constructor.  Takes an optional parameter of <code>configfile</code>,
the value of which should be the location of the XML configuration file.
The main benefit of passing the configfile parameter is that it enables
logging of actions performed with the XMLDocument object.

=head2 authentication()

  $self->authentication(client => "99000000", password => "mypassword");

Allows you to set the authentication data required in every transaction. This
method takes a hash of the allowed parameters, which are:

  * client (also known as your vTID)
  * password

These parameters will be supplied to you by DataCash when you sign up.

=head2 carddetails()

  $self->carddetails(pan => "343434343434343", expirydate => "03/09");

Allows you to set the parameters of a credit, debit or charge card. These may
include any of the following:

  * pan
  * expirydate
  * startdate
  * issuenumber
  * cpass

=head2 chequedetails()

  $self->chequedetails(sortcode => "826301", 
		       accountnumber => "01234567",
		       chequenumber => "000001",
		       type => "personal");

Allows you to set the details of a cheque, if you have cheque guarantee
facilities on your DataCash account.  Simply call the method as above, with
possible keys being:

  * sortcode
  * accountnumber
  * chequenumber
  * type (either "personal" or "business")

=head2 amount()

  $self->amount(amount => "313.37", currency => "USD");

  * amount
  * currency
  
If not supplied, the currency parameter will default to "GBP", indicating
that the transaction should be in Sterling.

=head2 method()

  $self->method("pre");

This is used to set the transaction type, with a single string being the only
required parameter.  The other possible options allowed are:

  * auth
  * pre
  * refund
  * fulfill
  * erp
  * cheque
  * txn_refund
  * fulfill

This method should be used after you have set at least some of the transaction 
details such as the card number, sort code etc as the <code>type</code> element
exists in numerous places in our Request schema.
 
=head2 reference()

  $self->reference("200111081343");

Almost all transactions sent to DataCash require a merchant reference to be sent
with the transaction.  As with the <code>method()</code> method, this needs to be
used after some of the transactional details have been set so that the reference
can be placed in the correct complex element in the XML document.

=head2 cv2()

  $self->cv2("519");

This method provides a simple way of inserting the cv2 number of a credit/debit card
into the XML document.

=head2 avs()

  $self->avs(street_address1 => '1 High Street',
	     street_address2 => 'Anytown',
	     postcode => 'AB11 2DE',
	     policy => '3');

This method allows you to insert the address details, plus optionally the AVS policy
that you wish this transaction to use.

Possible keys are:
 
  * street_address1
  * street_address2
  * street_address3
  * street_address4
  * postcode
  * policy

=head2 authcode()

  $self->authcode("157764");

This places the authcode into the XML document.  An authcode may need to be used for a 
variety of reasons, such as for fulfilling a "pre", and for processing a manual 
transaction.  For this reason, you should use this method after setting at least some
of the transactional details (e.g. card number or HistoricTxn details) so that the 
method knows where to put the authcode.

=head2 txntime()

Certain transactions may require the transaction time to be sent through, therefore this
method can be used to set the element in the document.

=head2 txndate()

Certain transactions may require the transaction date to be sent through, therefore this
method can be used to set the element in the document.

=head1 SEE ALSO

XML::Hash, DataCash::Config

=head1 AUTHOR

  * Kas Robson, DataCash Ltd
  * Kenny Scott, DataCash Ltd

