#!/usr/bin/perl

package XML::Hash;

use strict;
use Data::Dumper;

## $Format: "$XML::Hash::VERSION = \"$ProjectVersion$\";"$
$XML::Hash::VERSION = "0.45";

## Creator

sub new {
	my $type=shift;
	my %params=@_;

	my $self={
		xml=>{}
	};
	
	foreach my $param (keys %params) {
		if (grep /^$param$/,qw(config)) {
			$self->{$param}=$params{$param};
		} elsif ($param eq "logging") { ## Added for authd compatibility
			$self->{config}->{logging}=$params{$param};
		}
	}

	bless $self,$type;
	$self->{encoding} = "UTF-8";
	$self->_log(2,"created new instance");
	
	return $self;
}

## Accessor/mutator methods
## These are ultra-necessary, due to the fact that a lot of the hash
## structure is pretty much hidden, in order to handle attributes
## transparently... They're implemented separately, because otherwise
## setting things to undef gets real awkward.

sub set {
	my $self=shift;
	my $location=shift;
	my $value=shift;
	my %attrs=@_;

	# check for illegal characters
	make_legal(\$value);
	for my $key (keys %attrs) {
	  my $value = $attrs{$key};
	  make_legal(\$value);
	  $attrs{$key} = $value;
	}

	$self->_log(4,"setting $location to $value") if defined $value;

	if (%attrs) {
		$value={
			_data=>$value,
			_attributes=> {%attrs}
		};		
	}

	my @nodes=split /\./,$location;
	my $key=pop @nodes;
	
	my $base=$self->{xml};
	
	foreach my $node (@nodes) {
		my $tmp=$base;
	
		if (defined($tmp=$tmp->{$node})) {
			if (defined($tmp->{_data})) {
				$tmp=$tmp->{_data};
			}
			$base=$tmp;
		} else {
			$base->{$node}={};
			$base=$base->{$node};
		}
	}

	if (exists($base->{$key})) {
		if (ref($base->{$key})=~/ARRAY/) {
			push @{$base->{$key}},($value);
		} else {
			$base->{$key}=[$base->{$key},$value];
		}
	} elsif (exists($base->{_data})) {
		if (exists($base->{_data}->{$key})) {
			if (ref($base->{_data}->{$key})=~/ARRAY/) {
				push @{$base->{_data}->{$key}},($value);
			} else {
				$base->{_data}->{$key}=[$base->{_data}->{$key},$value];
			}
		} else {
			$base->{_data}->{$key}=$value;
		}
	} else {
		$base->{$key}=$value;
	}
}


## Method to change illegal XML characters into a safe representation

sub make_legal {
  my ($v) = @_;
  return if !defined $$v;
  $$v =~ s/\&(?!(amp|apos|quot|gt|lt)\;)/\&amp\;/g;
  $$v =~ s/\'/\&apos\;/g;
  $$v =~ s/\"/\&quot\;/g;
  $$v =~ s/\>/\&gt\;/g;
  $$v =~ s/\</\&lt\;/g;
}

## Method to determine - before doing a get() - just WTF the data retrieved
## is going to be... Returns "n", the number of elements if an array is being
## returned, 0 if the item is not an array, or undef if it doesn't exist.

sub is_array {
	my $self=shift;
	my $location=shift;
	my @nodes=split /\./,$location;
	my $k_index;
	my $key=pop @nodes;
	if ($key=~/^(\w+)\[(\d+)\]$/) {
		$key=$1;
		$k_index=$2-1;
	}
	
	my $base=$self->{xml};

	foreach my $node (@nodes) {
		my $index;
		if ($node=~/^(\w+)\[(\d+)\]$/) {
			$node=$1;
			$index=$2-1;
		}
	
		my $tmp=$base;
	
		if (defined($tmp=$tmp->{$node})) {
			if (((ref($tmp))=~/ARRAY/) and (defined($index))) {
				$base=@{$tmp}[$index];
				if (defined($base->{_data})) {
					$base=$base->{_data};
				}
			} else {
				$base=$tmp;
				if (defined($base->{_data})) {
					$base=$base->{_data};
				}
			}
		} else {
			return undef;
		}
	}

	if (exists($base->{$key})) {
		if (((ref($base->{$key}))=~/ARRAY/) and defined($k_index)) {
			return 0;
		} else {
			if ((ref($base->{$key}))=~/ARRAY/) {
				##print "140\n";
				return scalar(@{($base->{$key})});
			} else {
				##print "143\n";
				return 0;
			}
		}
	} elsif (exists($base->{_data})) {
		if (exists($base->{_data}->{$key})) {
			if ((ref($base->{_data}->{$key}))=~/ARRAY/) {
				return 0;
			} else {
				if ((ref($base->{$key}))=~/ARRAY/) {
					##print "156\n";
					return scalar(@{($base->{$key})});
				} else {
					##print "159\n";
					return 0;
				}
			}
		} else {
			##print "undef\n";
			##print "\n159\n";
			return undef;
		}
	} else {
		##print "undef\n";
		##print "\n164\n";
		return undef;
	}
}


## get() returns data (which may be a hashref, if the data is complex)
## followed by an anonymous hash of the attributes (if there are any)

sub get {
	my $self=shift;
	my $location=shift;

	my @nodes=split /\./,$location;
	my $k_index;
	my $key=pop @nodes;
	if ($key=~/^(\w+)\[(\d+)\]$/) {
		$key=$1;
		$k_index=$2-1;
	}
	
	my $base=$self->{xml};
	
	foreach my $node (@nodes) {
		my $index;
		if ($node=~/^(\w+)\[(\d+)\]$/) {
			$node=$1;
			$index=$2-1;
		}
	
		my $tmp=$base;
	
		if (defined($tmp=$tmp->{$node})) {
			if (((ref($tmp))=~/ARRAY/) and (defined($index))) {
				$base=@{$tmp}[$index];
				if (defined($base->{_data})) {
					$base=$base->{_data};
				}
			} else {
				$base=$tmp;
				if (defined($base->{_data})) {
					$base=$base->{_data};
				}
			}
		} else {
			return undef;
		}
	}

	if (exists($base->{$key})) {
		if (((ref($base->{$key}))=~/ARRAY/) and defined($k_index)) {
			return _getbits(@{$base->{$key}}[$k_index])
		} else {
			return _getbits($base->{$key});
		}
	} elsif (exists($base->{_data})) {
		if (exists($base->{_data}->{$key})) {
			if ((ref($base->{_data}->{$key}))=~/ARRAY/) {
				return _getbits(@{$base->{_data}->{$key}}[$k_index])
			} else {
				return _getbits($base->{_data}->{$key});
			}
		} else {
			return undef;
		}
	} else {
		return undef;
	}
}

sub _getbits {
	my $node=shift;
	
	if ((ref($node))=~/^HASH/) {
		if (exists $node->{_data}) {
			return ($node->{_data},$node->{_attributes});
		} else {
			return $node;
		}
	} elsif ((ref($node))=~/^ARRAY/) {
		if (grep {exists($_->{_data})} @{$node}) {
			my @array;
			foreach my $item (@{$node}) {
				push @array,([$item->{_data},$item->{_attributes}]);
			}
			return @array;
		} else {
			return $node;
		}
	} else {
		return $node;
	}
}


sub setencoding {
  
  my $self = shift;
  $self->{encoding} = shift;

}

sub getencoding {
  
  return shift->{encoding};

}

## Methods to convert between hash-structure and XML document. The
## implementation of these are slightly eccentric, coz the writexml()
## and readxml() methods are just wrappers round earlier (non O-O) subs...

sub writexml {
  
  my $self=shift;
  my $name=shift;
  my $encoding = (@_) ? shift : "UTF-8";
  my $xmldeclar='<?xml version="1.0" encoding="'.$self->{encoding}.'"?>'."\n";
  
  return($xmldeclar._xmlify($self->{xml},$name));
  
}


sub _xmlify {
	my $data=shift;
	$data = (defined $data) ? $data : "";

	my $name=shift || "XMLified";
	my $indent=shift || 0;
	
	if ($name=~/^(\w+)\[(\w+)\]$/) {
		$name=$1;
	}
	
	my $document=undef;
	
	if ((ref $data)=~/^HASH/) {

		if (exists $data->{_data}) {

			if (exists $data->{_attributes}) {
				$document .= ("  "x$indent);
				$document .= ("<".$name._attributes($data->{_attributes}).">");
				if (defined $data->{_data}) {
				  if ((ref($data->{_data}))=~/^HASH/) {
				    $document .= "\n";
				    foreach my $element (grep !/^_/,(keys %{$data->{_data}})) {
				      $document .= _xmlify($data->{_data}->{$element},$element,$indent+1);
				      ##$document .= "\n";
				    }
				    $document .= ("  "x$indent);
				  } 
				  else {
				    $document .= $data->{_data};
				  }
				}
				$document .= ("</".$name.">\n");
			} else {
				$document .= ("  "x$indent);
				$document .= ("<".$name.">".$data->{_data}."</".$name.">\n");
			}

		} else {

			if (exists $data->{_attributes}) {
				$document .= ("  "x$indent);
				$document .= ("<".$name._attributes($data->{_attributes}).">\n");
			} else {
				$document .= ("  "x$indent);
				$document .= ("<".$name.">\n");
			}
			
			foreach my $element (grep !/^_/,(keys %{$data})) {
				$document .= _xmlify($data->{$element},$element,$indent+1);
			}
			$document .= ("  "x$indent);
			$document .= "</".$name.">\n";
		}

	} elsif ((ref $data)=~/^ARRAY/) {

		foreach my $element (@{$data}) {
			$document .= _xmlify($element,$name,$indent);
		}

	} else {
		$document .= ("  "x$indent);
		$document .= ("<".$name.">".$data."</".$name.">\n");
	}
	
	return $document;
}

sub _attributes {
  my $hash=shift;
	
  my $attrstring="";
	
  foreach my $attr (keys %{$hash}) {
    my $value = ( defined $hash->{$attr} ) ? $hash->{$attr} : '';
    $attrstring .= " $attr='$value'";
  }
	
  return $attrstring;
}


sub readxml {
	my $self=shift;
	my $xml=shift;
	my $length=length($xml);
	
	$self->_log(3,"reading $length bytes of XML");
	
	$self->{xml}=_hashify($xml);
}

sub _hashify {
	my $excerpt=shift;
	
	$excerpt=~s/\n//g;
	
	## Get rid of the likes of...
	##    <?xml version="1.0" encoding="UTF-8"?>
	$excerpt=~s/<\?.*?\?>//g;


	if ($excerpt=~s/^\s*<(\w+)(\s+([^<>]*?))?\s*\/>(.*)/$4/) { ## empty

		my $tag=$1;
		my $attrs=$3;
		my $leftovers=$4;
		
		##debug("Tag: $tag","Attrs: $attrs","Leftovers: $leftovers");	
		
		my $tree={};

		if ($attrs) {
			$tree->{$tag}->{_data}='';
			$tree->{$tag}->{_attributes}=_attrhash($attrs);
		} else {
			$tree->{$tag}='';
		}

		if ($leftovers) {
			$tree=_merge($tree,_hashify($leftovers));	
		}

		return $tree;		
	
	} elsif ($excerpt=~s/^\s*<(\w+)(\s+([^<>]*?))?>(.*?)<\/\1>(.*)/$5/) { ## non-empty

		my $tag=$1;
		my $attrs=$3;
		my $content=$4;
		my $leftovers=$5;

		##debug("Tag: $tag","Attrs: $attrs","Content: $content","Leftovers: $leftovers");

		my $tree={};

		if ($attrs) {
			$tree->{$tag}->{_data}=_hashify($content);
			$tree->{$tag}->{_attributes}=_attrhash($attrs);
		} else {
			$tree->{$tag}=_hashify($content);
		}

		if ($leftovers !~ /^\s*$/) {
			$tree=_merge($tree,_hashify($leftovers));	
		}

		return $tree;

	} elsif ($excerpt=~/^\s*$/) { ## whitespace or nada
		##debug("Whitespace or empty");
		return '';
	} else { ## Must be chars
		##debug("Just some chars");
		return $excerpt;
	}	
	
}

sub _attrhash {
	my $attrlist=shift;
	my $attrhash=shift;

	##debug("Hashifying attributes");

	while (not ($attrlist=~/^\s*$/)) {
		##debug($attrlist);
		if ($attrlist=~s/([\w-]+?)\s*?=\s*?['"]([^'"]*)['"]\s*?(.*)/$3/) {
			$attrhash->{$1}=$2;
		} else {
			##debug("Attributes are bogus (probably not quoted)");
			return undef;
		}
	}
	
	return $attrhash;
}


sub _merge {
	my $tree_a=shift;
	my $tree_b=shift;
	
	foreach my $key (keys %{$tree_b}) {
		if (exists $tree_a->{$key}) {
			if (ref($tree_b->{$key})=~/ARRAY/) {
				push @{$tree_b->{$key}},($tree_a->{$key});
				$tree_a->{$key}=$tree_b->{$key};
			} else {
				$tree_a->{$key}=[$tree_a->{$key},$tree_b->{$key}];
			}
		} else {
			$tree_a->{$key}=$tree_b->{$key};
		}
	}
	
	return $tree_a;
}


## Utility methods

sub _log {
	my $self=shift;
	my $whoami=ref($self);
	my $lvl=shift;
	
	if (defined $self->{config} && defined $self->{config}->{logging} && $self->{config}->{logging} >= $lvl) {
		foreach my $msg (@_) {
		
			if ($msg=~/^(setting (.*) to )(.*)$/) {
				my $stem=$1;
				my $node=$2;
				my $value=$3;
				
				if (grep /$node/,@{$self->{config}->{obscure}}) {
					$msg=$stem._mask($3);
				}
			}
		
			if ((ref($self->{config}->{logfh})) eq "FileHandle") {
				my $fh=$self->{config}->{logfh};
				print $fh scalar(localtime(time))," [$$] $whoami $msg\n";
			} else {
				print STDERR "$msg\n";
			}
		}
	}
}

sub _mask {
	my $data=shift;
	return("[".("X"x(length($data)))."]");
}

1;

######################################################################!#########
__END__

=head1 NAME

XML::Hash

=head1 SYNOPSIS

   use XML::Hash;
   my $xmlhash=new XML::Hash(debug=>1);
   ## do some stuff to populate the object...
   print $xmlhash->writexml();

=head1 DESCRIPTION

XML::Hash provides a means of converting data back and forth between an
XML representation and representation internally as a big hashref. This
version has been given a facelift, including (most significantly) an
object-oriented interface.

=head1 PUBLIC METHODS

=head2 new()

The creator method, new(), allows you to create a new instance of the
XML::Hash object. It can be passed a hash of configuration options, such
as:

  * config
  * logging
  
Normally, "config" is used to pass a DataCash::Config object around in
the API. For use in other scenarios, "logging" can be used to set the
verbosity of XML::Hash output to logs. In this case, log output will
just be written to STDERR.
  
=head2 writexml()

Once the XML::Hash has been populated, either manually or by using the
accessor/mutator methods provided, calling the writexml() method will
return the XML document represented by the object.

For users of earlier versions of XML::Hash, this is an object-oriented
wrapper around the XMLify() function.

=head2 readxml()

As an alternative means of populating the object, the readxml() method
allows you to pass a string containing an XML document, which will be
parsed and stored in the hash internally.

For users of earlier versions of XML::Hash, this is an object-oriented
wrapper around the hashify() function.

=head2 set()

This method allows you to set data at an arbitrary point in the internal
hash structure representing the XML document. If the specified node does
not exist, it will be created.

In its simplest form, the method is used thusly. This example would create
the element <foo>bar</foo> within the document.

   my $foo=new XML::Hash;
   $foo->set("foo","bar");

The set() method can also use a (sort of) object-oriented syntax to allow
you to access deeply nested elements. To create <foo><bar>baz</bar></foo>:

   my $foo=new XML::Hash;
   $foo->set("foo.bar","baz");

Finally, a hash of attributes can be passed with the element definition.

   my $foo=new XML::Hash;
   $foo->set("foo.bar","baz",color=>'blue');

This example would create <foo><bar color='blue'>baz</bar></foo>.

When setting multiple instances of an unbounded element, an index should be
passed in the location, in order to know which instance is being set.

   $foo->set("foo.bar[1].baz","123456");
   $foo->set("foo.bar[2].baz","654321");

This would create...

   <foo>
      <bar>
         <baz>123456</baz>
      </bar>
      <bar>
         <baz>654321</baz>
      </bar>
   </foo>

There is one final complication, when setting attributes on non-leaf nodes.
This is most easily achieved in two stages. For example, if you were
attempting to create the excerpt of XML <foo time='19:30'><bar>baz</bar></foo>,
you would need to use code along the lines of the following:

   my $foo=new XML::Hash;
   $foo->set("foo",undef,time=>'19:30');
   $foo->set("foo.bar","baz");

The undef will be effectively overwritten. This is, however, the only case
where exisiting values will be overwritten. Where duplicate keys are
encountered, the values will be pushed onto a list, allowing you to create
unbounded elements. Non-duplicate keys will allow you to create "parallel"
elements, such as <foo><bar>baz</bar><qux>qub</qux></foo>, as created by the
code:

   my $foo=new XML::Hash;
   $foo->set("foo.bar","baz");
   $foo->set("foo.qux","qub");

=head2 get()

The get() method provides access to elements within the internally represented
hash structure, using a similar O-O syntax to that employed by the set() method.

   my $foo=new XML::Hash;
   ## do stuff to $foo
   my $thingy=$foo->get("foo.bar");

In addition, get supports the same index notation as the set() method.
   
What get() returns depends a little on the element itself. If the element is not
defined, get() will return undef. Assuming it is present, however, the method will
return the contents of the element (which may be a hashref, if the data is complex)
followed by an anonymous hash of the attributes (if there are any).

=head2 is_array()

Method to determine - before doing a get() - just WTF the data retrieved
is going to be... Returns I<n>, the number of elements if an array is being
returned, 0 if the item is not an array, or undef if it doesn't exist.


=head1 INTERNALS

The internals of XML::Hash remain virtually unchanged from previous versions. The
data I<has>, however, been moved into XML::Hash::xml, in order to allow the
XML::Hash object to store other parameters (such os mode of operation optons, etc).

The hash structure used to represent XML elements is as follows.

     element => {
     	first_child_elem => {
     		_attributes=> {
     			attribute1 => "value1",
     			attribute2 => "value2"
     		},
     		_data => "data contained in tag"
     	},
     	second_child_elem => "foo",
     	third_child_elem => ["foo","bar","baz"]
     }
     
This would be rendered as:

     <element>
     	<first_child_elem attribute1='value1' attribute2='value2'>
     		data contained in tag
     	</first_child_elem>
     	<second_child_elem>foo</second_child_elem>
     	<third_child_elem>foo</third_child_elem>
     	<third_child_elem>bar</third_child_elem>
     	<third_child_elem>baz</third_child_elem>
     </element>
     
Where an element has attributes, the _attributes and _data keys should be used.
Otherwise, they can be shortcut.

See the test-scripts t/*.pl for further examples.

=head1 CHANGELOG

=head2 0.31

  * Modification to _hashify(), which is used internally by readxml(),
    in order to set empty elements to the empty-string, rather than
    to undef.

=head2 0.30

  * Bugfix to readxml(), so that it eliminates empty elements first.

=head2 0.29

  * Added the is_array() method to supplement get().
  * Fixed a gnarly bug in get() that meant traversal didn't always find
    _data elements with the internal representation.
  * Eradicated blank lines that appear in some odd cases involving _data
    elements in internal representation.

=head2 0.28

  * Further modification to get() which *will* now pass an array back
    if not told to do otherwise...

=head2 0.27

  * Modification to get(), so that it now supports indices in the
    same form as set(). These are numbered from 1, NOT ZERO, at the
    moment, though this can be changed.

=head2 0.26

  * Modification to get() and related methods, so that get()'ing an
    unbounded element will return in the form...
    
       ((item1data,item1attributes),(item2data,item2attributes),...)

  * Some tidying up of the embedded POD documentation.
  * Removal of a stray print in _merge(), used in debugging a previous
    issue.

=head2 0.25

  * Further modifications to set() method, to deal with insertion of new
    sub-elements on an element with attributes and data.

=head2 0.24

  * Modification to set() method for better handling of attributes [amr].

=head2 0.23

  * Further fix for complex unbounded elements - this time to the _merge(),
    which is used as part of the readxml() method.

=head2 0.22

  * Fix to allow complex and unbounded elements. This is achieved through
    further augmentation (ie. complication) of the memory structure used
    internally to represent the XML document.

=head2 0.21

  * Fixed Makefile.pl

=head2 0.20

  * Housekeeping release

=head2 0.19

  * Fix to allow attribute names with hyphens in them.

=head2 0.18

  * Fix to attribute handling in readxml() method.

=head2 0.17

  * Added _log() private method, with support for logfile handle
  * Added "logging" constructor parameter, to make XML::Hash more
    compatible in environments other than the DataCash4 API.

=head2 0.16

  * Fixed get() method for elements of complex type

=head2 0.15

  * Added MakeMaker support.
  * Added support for XML declarations of the form...

       <?xml version="1.0" encoding="UTF-8"?>

=head2 0.13

  * Housekeeping release, with fixes to documentation and example code.

=head2 0.12

  * This version is pretty much a complete rewrite. XML::Hash now has an
    object-oriented interface, which (hopefully) makes it more powerful
    and easier to use. As such, it is (obviously) not 100% compatible with
    earlier versions... Hey, this is alpha software, okay.

=head2 0.11

  * Fixed bug where malformed attributes can lead to infinite regression.

=head2 0.10

  * PRCS keyword versioning added.
  * Debug facilities added.

=head2 0.3

  * Perldoc added.
  * First "stable" version.

=head1 KNOWN BUGS/TO-DO LIST

  * Remove use of Data::Dumper prior to release.
  * Apply similar mangling to get() as I have done to set() to allow
    for greater control of unbounded complex elements. The same magic
    should be done by readxml().
  * Make use versus non-use of indexed unboundeds more consistent.
  * Remove all uses of lists in the internal representation, as the
    mangling on get(), set() and readxml() will make it unnecessary.
  * Investigate possible small inconsistency between readxml() and
    writexml() methods.
  * Make debug more useful for non O-O subs.
  * One day, I'd really like to rewrite all the tree traversal stuff.
  * Add a delete() method, in the style of set() and get().
 
=head1 AUTHOR

Kas Robson (kxr@datacash.com), DataCash Ltd.

=head1 LICENSE
